# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Records relations tests helpers."""

from __future__ import unicode_literals

import json

from flask import url_for

from invenio_app_ils.records_relations.api import RecordRelationsExtraMetadata

TYPES_ENDPOINTS = {
    "relation": {
        "docid": "invenio_app_ils_relations.docid_relations",
        "serid": "invenio_app_ils_relations.serid_relations",
    },
    "get": {
        "docid": "invenio_records_rest.docid_item",
        "serid": "invenio_records_rest.serid_item",
    },
}


def recrel_fetch_record(client, json_headers, url):
    """Fetch a record."""
    res = client.get(url_for(url[0], pid_value=url[1]), headers=json_headers)
    assert res.status_code == 200
    return json.loads(res.data.decode("utf-8"))["metadata"]


def recrel_assert_record_relations(record, expected):
    """Test extra metadata and `relations` field on parent record."""
    extra_metadata_field_name = RecordRelationsExtraMetadata.field_name()
    relations_metadata = record.get(extra_metadata_field_name, {})
    expected_metadata = expected.get(extra_metadata_field_name, {})

    relations = record.get("relations", {})
    expected_relations = expected.get("relations", {})

    assert relations_metadata.keys() == expected_metadata.keys()
    for relation, records in relations_metadata.items():
        assert len(records) == len(expected_metadata[relation])
        for rec in records:
            assert rec in expected_metadata[relation]
    assert relations.keys() == expected_relations.keys()
    for relation, records in relations.items():
        assert len(records) == len(expected_relations[relation])
        for rec in records:
            assert rec in expected_relations[relation]


def recrel_do_request_for_valid_relation(
    client, json_headers, payload, url, method="POST", expect_status_code=None
):
    """Create a new valid relation."""
    func = client.delete if method == "DELETE" else client.post
    res = func(
        url_for(url[0], pid_value=url[1]),
        headers=json_headers,
        data=json.dumps(payload),
    )
    if expect_status_code is None:
        expected_status_code = 200 if method == "DELETE" else 201
    else:
        expected_status_code = expect_status_code
    assert res.status_code == expected_status_code
    if res.status_code not in [200, 201]:
        return
    return json.loads(res.data.decode("utf-8"))["metadata"]


def recrel_choose_endpoints_and_do_request(
    client_params,
    relation,
    payload,
    create_using_pid1=True,
    expect_status_code=None,
):
    """Choose which endpoint to use, create relation, return records."""
    client, json_headers, method = client_params
    pid1, pid1_type, pid2, pid2_type = relation

    if create_using_pid1:
        url_create_rel = (TYPES_ENDPOINTS["relation"][pid1_type], pid1)
        url_other = (TYPES_ENDPOINTS["get"][pid2_type], pid2)

        record1 = recrel_do_request_for_valid_relation(
            client,
            json_headers,
            payload,
            url_create_rel,
            method=method,
            expect_status_code=expect_status_code,
        )
        record2 = recrel_fetch_record(client, json_headers, url_other)
    else:
        url_create_rel = (TYPES_ENDPOINTS["relation"][pid2_type], pid2)
        url_other = (TYPES_ENDPOINTS["get"][pid1_type], pid1)

        record2 = recrel_do_request_for_valid_relation(
            client,
            json_headers,
            payload,
            url_create_rel,
            method=method,
            expect_status_code=expect_status_code,
        )
        record1 = recrel_fetch_record(client, json_headers, url_other)
    return record1, record2



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Order schema for marshmallow loader."""

from invenio_circulation.records.loaders.schemas.json import DateString
from invenio_records_rest.schemas import RecordMetadataSchemaJSONV1
from marshmallow import EXCLUDE, Schema, fields, pre_load, validate

from invenio_app_ils.acquisition.api import Order
from invenio_app_ils.records.loaders.schemas.changed_by import (
    ChangedBySchema,
    set_changed_by,
)
from invenio_app_ils.records.loaders.schemas.price import PriceSchema


class OrderLineSchema(Schema):
    """Order line schema."""

    class Meta:
        """Meta attributes for the schema."""

        unknown = EXCLUDE

    budget_code = fields.Str()
    copies_ordered = fields.Int(required=True)
    copies_received = fields.Int()
    document_pid = fields.Str(required=True)
    inter_departmental_transaction_id = fields.Str()
    is_donation = fields.Bool()
    is_patron_suggestion = fields.Bool()
    medium = fields.Str(required=True)
    notes = fields.Str()
    patron_pid = fields.Str()
    payment_mode = fields.Str()
    purchase_type = fields.Str()
    recipient = fields.Str(required=True)
    total_price = fields.Nested(PriceSchema)
    unit_price = fields.Nested(PriceSchema)


class PaymentSchema(Schema):
    """Payment schema."""

    class Meta:
        """Meta attributes for the schema."""

        unknown = EXCLUDE

    debit_cost = fields.Nested(PriceSchema)
    debit_cost_main_currency = fields.Nested(PriceSchema)
    debit_date = DateString()
    debit_note = fields.Str()
    internal_purchase_requisition_id = fields.Str()
    mode = fields.Str()


class OrderSchemaV1(RecordMetadataSchemaJSONV1):
    """Order schema."""

    class Meta:
        """Meta attributes for the schema."""

        unknown = EXCLUDE

    cancel_reason = fields.Str()
    created_by = fields.Nested(ChangedBySchema)
    expected_delivery_date = DateString()
    funds = fields.List(fields.Str())
    grand_total = fields.Nested(PriceSchema)
    grand_total_main_currency = fields.Nested(PriceSchema)
    legacy_id = fields.Str()
    notes = fields.Str()
    order_date = DateString()
    order_lines = fields.List(fields.Nested(OrderLineSchema), required=True)
    payment = fields.Nested(PaymentSchema)
    received_date = DateString()
    status = fields.Str(required=True, validate=validate.OneOf(Order.STATUSES))
    updated_by = fields.Nested(ChangedBySchema)
    provider_pid = fields.Str(required=True)

    @pre_load
    def set_changed_by(self, data, **kwargs):
        """Automatically set `created_by` and `updated_by`."""
        record = self.context.get("record")
        return set_changed_by(data, record)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Notifications tasks."""

import json

from celery import shared_task
from flask import current_app

from invenio_app_ils.notifications.models import NotificationsLogs


def _log_notification_to_db(data, log_message):
    """Log notification in DB."""
    data_dict = dict(
        action=data.get("action", None),
        pid_type=data.get("pid_type", None),
        pid_value=data.get("pid_value", None),
        recipient_user_id=data["patron_id"],
        is_manually_triggered=data["is_manually_triggered"],
        message_id=data["id"],
        send_log=log_message,
    )
    NotificationsLogs.create(data_dict)


@shared_task
def log_successful_notification(_, data):
    """Log successful notification task."""
    log_msg = dict(
        name="notification",
        action="success",
        message_id=data["id"],
        data=data,
    )
    current_app.logger.info(json.dumps(log_msg, sort_keys=True))
    _log_notification_to_db(data, "Success")


@shared_task
def log_error_notification(request, exc, traceback, data, **kwargs):
    """Log error when sending notification task."""
    error = dict(
        name="notification",
        action="error",
        message_id=data["id"],
        task_id=request.get("id"),
        task=request.get("task"),
        exception=repr(exc),
        data=data,
    )
    current_app.logger.exception(json.dumps(error, sort_keys=True), exc_info=exc)
    _log_notification_to_db(data, "Error: " + repr(exc))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test anonymization of users."""

from copy import deepcopy

import pytest
from flask import url_for
from invenio_accounts.models import User
from invenio_circulation.api import Loan
from invenio_circulation.proxies import current_circulation
from invenio_db import db
from invenio_oauthclient.models import RemoteAccount, UserIdentity
from invenio_search import current_search
from invenio_userprofiles.models import UserProfile

from invenio_app_ils.acquisition.api import Order
from invenio_app_ils.acquisition.search import OrderSearch
from invenio_app_ils.circulation.search import (
    get_active_loans_by_patron_pid,
    get_loans_by_patron_pid,
)
from invenio_app_ils.document_requests.api import DocumentRequest
from invenio_app_ils.document_requests.search import DocumentRequestSearch
from invenio_app_ils.errors import AnonymizationActiveLoansError
from invenio_app_ils.ill.api import BorrowingRequest
from invenio_app_ils.ill.search import BorrowingRequestsSearch
from invenio_app_ils.patrons.anonymization import anonymize_patron_data
from invenio_app_ils.patrons.api import patron_exists
from tests.helpers import user_login, user_logout


def check_user_deleted(user_pid):
    """Check if user was deleted from database."""
    with db.session.begin_nested():
        user_identity = UserIdentity.query.filter(
            UserIdentity.id_user == user_pid
        ).first()

        assert user_identity is None

        remote_account = RemoteAccount.query.filter(
            RemoteAccount.user_id == user_pid
        ).first()
        assert remote_account is None

        user_profile = UserProfile.query.filter(UserProfile.user_id == user_pid).first()
        assert user_profile is None

        user = User.query.filter(User.id == user_pid).first()
        assert user is None


def check_user_exists(user_pid):
    """Check if user exists in database."""
    with db.session.begin_nested():
        assert patron_exists(user_pid)


def check_user_activity(app, user_pid, client, json_headers):
    """Check if there are records related to the user."""
    # wait ES refresh
    current_search.flush_and_refresh(index="*")

    AnonymousPatron = app.config["ILS_PATRON_ANONYMOUS_CLASS"]
    anonymous_patron_fields = AnonymousPatron().dumps_loader()

    loans = get_loans_by_patron_pid(user_pid).scan()
    for hit in loans:
        # test ES
        assert hit["patron"] == anonymous_patron_fields
        # test DB
        loan = Loan.get_record_by_pid(hit.pid)
        assert loan["patron"] == anonymous_patron_fields
        # test REST
        url = url_for("invenio_records_rest.loanid_item", pid_value=hit.pid)
        res = client.get(url, headers=json_headers)
        assert res.get_json()["metadata"]["patron"] == anonymous_patron_fields

    borrowing_requests = BorrowingRequestsSearch().search_by_patron_pid(user_pid).scan()
    for hit in borrowing_requests:
        # test ES
        assert hit["patron"] == anonymous_patron_fields
        # test DB
        borrowing_request = BorrowingRequest.get_record_by_pid(hit.pid)
        assert borrowing_request["patron"] == anonymous_patron_fields
        # test REST
        url = url_for("invenio_records_rest.illbid_item", pid_value=hit.pid)
        res = client.get(url, headers=json_headers)
        assert res.get_json()["metadata"]["patron"] == anonymous_patron_fields

    document_requests = DocumentRequestSearch().search_by_patron_pid(user_pid).scan()
    for hit in document_requests:
        # test ES
        assert hit["patron"] == anonymous_patron_fields
        # test DB
        document_request = DocumentRequest.get_record_by_pid(hit.pid)
        assert document_request["patron"] == anonymous_patron_fields
        # test REST
        url = url_for("invenio_records_rest.dreqid_item", pid_value=hit.pid)
        res = client.get(url, headers=json_headers)
        assert res.get_json()["metadata"]["patron"] == anonymous_patron_fields

    acquisitions = OrderSearch().search_by_patron_pid(user_pid).scan()
    for hit in acquisitions:
        # test ES
        assert hit["patron"] == anonymous_patron_fields
        # test DB
        acquisition = Order.get_record_by_pid(hit.pid)
        assert acquisition["patron"] == anonymous_patron_fields
        # test REST
        url = url_for("invenio_records_rest.acqoid_item", pid_value=hit.pid)
        res = client.get(url, headers=json_headers)
        assert res.get_json()["metadata"]["patron"] == anonymous_patron_fields


def cancel_active_loans(patron_pid, client, users):
    """Cancel active loans of a patron."""
    user = user_login(client, "admin", users)

    active_loans = get_active_loans_by_patron_pid(patron_pid).scan()

    for hit in active_loans:
        loan = Loan.get_record_by_pid(hit.pid)
        params = deepcopy(loan)
        params.update(
            dict(
                cancel_reason="Loan cancelled to anonymize user.",
                transaction_user_pid=str(user.id),
            )
        )
        current_circulation.circulation.trigger(loan, **dict(params, trigger="cancel"))

        loan.commit()
        db.session.commit()
        current_circulation.loan_indexer().index(loan)

    current_search.flush_and_refresh(index="*")


def test_anonymization(app, client, json_headers, users, testdata):
    """Test anonymization of a user."""
    # Anonymize patron
    patron_pid = users["patron3"].id

    check_user_exists(patron_pid)
    cancel_active_loans(patron_pid, client, users)
    anonymize_patron_data(patron_pid)
    check_user_deleted(patron_pid)

    user_login(client, "admin", users)
    check_user_activity(app, patron_pid, client, json_headers)
    user_logout(client)

    # Anonymize librarian
    librarian_pid = users["librarian"].id

    check_user_exists(librarian_pid)
    cancel_active_loans(librarian_pid, client, users)
    anonymize_patron_data(librarian_pid)
    check_user_deleted(librarian_pid)

    # Anonymize patron while logged in
    patron_pid = users["patron2"].id
    user_login(client, "patron2", users)

    check_user_exists(patron_pid)
    cancel_active_loans(patron_pid, client, users)
    anonymize_patron_data(patron_pid)
    check_user_deleted(patron_pid)

    # It should fail when anonymizing patron with active loans
    patron_pid = users["patron1"].id
    check_user_exists(patron_pid)
    with pytest.raises(AnonymizationActiveLoansError):
        assert anonymize_patron_data(patron_pid)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2021 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test internal locations."""

import json

import pytest
from flask import url_for

from invenio_app_ils.errors import LocationNotFoundError
from invenio_app_ils.internal_locations.api import InternalLocation
from tests.helpers import user_login

_HTTP_OK = [200, 201, 204]
_INTERNAL_LOCATION_PID = "ilocid-1"
_INTERNAL_LOCATION_NAME = "A location"
_ITEM_ENDPOINT = "invenio_records_rest.ilocid_item"
_LIST_ENDPOINT = "invenio_records_rest.ilocid_list"


def test_internal_locations_validation(db, testdata):
    """Test validation when updating an Internal Location."""
    intloc_pid = testdata["internal_locations"][0]["pid"]
    intloc = InternalLocation.get_record_by_pid(intloc_pid)

    intloc["location_pid"] = "not_found_pid"
    with pytest.raises(LocationNotFoundError):
        intloc.commit()


def test_internal_locations_permissions(client, testdata, json_headers, users):
    """Test internal locations endpoints permissions."""
    dummy_internal_location = dict(
        location_pid=testdata["locations"][0]["pid"],
        name=_INTERNAL_LOCATION_NAME,
    )
    tests = [
        ("admin", _HTTP_OK, dummy_internal_location),
        ("librarian", _HTTP_OK, dummy_internal_location),
        ("patron1", [403], dummy_internal_location),
        ("anonymous", [401], dummy_internal_location),
    ]

    def _test_list(expected_status):
        """Test get list."""
        url = url_for(_LIST_ENDPOINT)
        res = client.get(url, headers=json_headers)
        assert res.status_code in expected_status

    def _test_create(expected_status, data):
        """Test record creation."""
        url = url_for(_LIST_ENDPOINT)
        res = client.post(url, headers=json_headers, data=json.dumps(data))
        assert res.status_code in expected_status

        if res.status_code < 400:
            record = res.get_json()["metadata"]
            assert record["name"] == _INTERNAL_LOCATION_NAME
            return record["pid"]

    def _test_update(expected_status, data, pid):
        """Test record update."""
        pid_value = pid or _INTERNAL_LOCATION_PID
        url = url_for(_ITEM_ENDPOINT, pid_value=pid_value)
        res = client.put(url, headers=json_headers, data=json.dumps(data))
        assert res.status_code in expected_status
        if res.status_code < 400:
            record = res.get_json()["metadata"]
            assert record["name"] == _INTERNAL_LOCATION_NAME

    def _test_read(expected_status, pid):
        """Test record read."""
        pid_value = pid or _INTERNAL_LOCATION_PID
        url = url_for(_ITEM_ENDPOINT, pid_value=pid_value)
        res = client.get(url, headers=json_headers)
        assert res.status_code in expected_status

    def _test_delete(expected_status, pid):
        """Test record delete."""
        pid_value = pid or _INTERNAL_LOCATION_PID
        url = url_for(_ITEM_ENDPOINT, pid_value=pid_value)
        res = client.delete(url, headers=json_headers)
        assert res.status_code in expected_status

    for username, expected_status, data in tests:
        user_login(client, username, users)
        _test_list(expected_status)
        pid = _test_create(expected_status, data)
        _test_update(expected_status, data, pid)
        _test_read(expected_status, pid)
        _test_delete(expected_status, pid)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio App ILS circulation loaders."""

from invenio_app_ils.records.loaders import ils_marshmallow_loader

from .schemas.json.bulk_extend import BulkExtendLoansSchemaV1
from .schemas.json.loan_checkout import LoanCheckoutSchemaV1
from .schemas.json.loan_request import LoanRequestSchemaV1
from .schemas.json.loan_self_checkout import LoanSelfCheckoutSchemaV1
from .schemas.json.loan_update_dates import LoanUpdateDatesSchemaV1

loan_request_loader = ils_marshmallow_loader(LoanRequestSchemaV1)
loan_checkout_loader = ils_marshmallow_loader(LoanCheckoutSchemaV1)
loan_self_checkout_loader = ils_marshmallow_loader(LoanSelfCheckoutSchemaV1)
loan_update_dates_loader = ils_marshmallow_loader(LoanUpdateDatesSchemaV1)
loans_bulk_update_loader = ils_marshmallow_loader(BulkExtendLoansSchemaV1)



# Copyright (C) 2021 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Notifications views."""

from flask import Blueprint, abort
from sqlalchemy.orm.exc import NoResultFound
from webargs import fields, validate
from webargs.flaskparser import use_kwargs

from invenio_app_ils.notifications.models import NotificationsLogs
from invenio_app_ils.permissions import need_permissions


def serialize_notification(notification):
    """Serialize notification from db."""
    if notification is None:
        return {}
    notification.triggered_on = notification.triggered_on.isoformat()
    notification = notification.__dict__
    del notification["_sa_instance_state"]
    return notification


def get_notifications_blueprint(_):
    """Create the notifications blueprint."""
    blueprint = Blueprint("invenio_app_ils_notifications", __name__)

    @blueprint.route("/notifications")
    @use_kwargs(
        {
            "recipient_user_id": fields.Int(),
            "pid_type": fields.Str(),
            "pid_value": fields.Str(),
            "action": fields.Str(),
            "size": fields.Int(validate=[validate.Range(min=1, max=100)]),
        }
    )
    @need_permissions("send-notification-to-patron")
    def get_notifications(
        recipient_user_id=None,
        pid_type=None,
        pid_value=None,
        action=None,
        size=20,
    ):
        filters = {}
        if recipient_user_id:
            filters.update({"recipient_user_id": recipient_user_id})
        if pid_type:
            filters.update({"pid_type": pid_type})
        if pid_value:
            filters.update({"pid_value": pid_value})
        if action:
            filters.update({"action": action})

        notifications = (
            NotificationsLogs.query.filter_by(**filters)
            .order_by(NotificationsLogs.triggered_on.desc())
            .limit(size)
            .all()
        )

        for i, notification in enumerate(notifications):
            notifications[i] = serialize_notification(notification)

        return {"hits": notifications, "total": len(notifications)}

    @blueprint.route("/notifications/<int:id>")
    @need_permissions("send-notification-to-patron")
    def get_notification(id):
        try:
            notification = NotificationsLogs.query.filter_by(id=id).one()
            return serialize_notification(notification)
        except NoResultFound:
            abort(404)

    return blueprint



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Simple test of version import."""

from __future__ import absolute_import, print_function


def test_version():
    """Test version import."""
    from invenio_app_ils import __version__

    assert __version__



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test accessibility of document request resource endpoints."""

from flask import url_for

from tests.helpers import user_login, validate_response


def test_get_document_request_endpoint(client, json_headers, testdata, users):
    """Test GET permissions."""

    tests = [
        ("patron1", "dreq-1", 200),
        ("librarian", "dreq-1", 200),
        ("admin", "dreq-1", 200),
        ("patron2", "dreq-1", 403),
        ("anonymous", "dreq-1", 401),
        ("patron2", "dreq-2", 200),
        ("librarian", "dreq-2", 200),
        ("admin", "dreq-2", 200),
        ("patron1", "dreq-2", 403),
        ("anonymous", "dreq-2", 401),
    ]

    for user, res_id, expected_resp_code in tests:
        user_login(client, user, users)
        url = url_for("invenio_records_rest.dreqid_item", pid_value=res_id)
        validate_response(client, "get", url, json_headers, None, expected_resp_code)


def test_document_request_add_document(client, json_headers, testdata, users):
    """Test add document to Document Request permissions."""

    tests = [
        ("patron1", "dreq-1", 403),
        ("librarian", "dreq-1", 202),
        ("admin", "dreq-1", 202),
        ("anonymous", "dreq-1", 401),
    ]

    for user, res_id, expected_resp_code in tests:
        user_login(client, user, users)
        url = url_for(
            "invenio_app_ils_document_requests.dreqid_document",
            pid_value=res_id,
        )
        data = {"document_pid": "docid-1"}
        validate_response(client, "post", url, json_headers, data, expected_resp_code)


def test_document_request_remove_document(client, json_headers, testdata, users):
    """Test remove document from Document Request permissions."""

    tests = [
        ("patron1", "dreq-1", 403),
        ("librarian", "dreq-1", 202),
        ("admin", "dreq-1", 202),
        ("anonymous", "dreq-1", 401),
    ]

    for user, res_id, expected_resp_code in tests:
        user_login(client, user, users)
        url = url_for(
            "invenio_app_ils_document_requests.dreqid_document",
            pid_value=res_id,
        )
        data = {"document_pid": "docid-1"}
        validate_response(client, "delete", url, json_headers, data, expected_resp_code)


def test_document_request_add_provider(client, json_headers, testdata, users):
    """Test add provider to Document Request permissions."""

    tests = [
        ("patron1", "dreq-1", 403),
        ("librarian", "dreq-1", 202),
        ("admin", "dreq-1", 202),
        ("anonymous", "dreq-1", 401),
    ]

    for user, res_id, expected_resp_code in tests:
        user_login(client, user, users)
        url = url_for(
            "invenio_app_ils_document_requests.dreqid_provider",
            pid_value=res_id,
        )
        data = {
            "physical_item_provider": {
                "pid": "acqoid-1",
                "pid_type": "acqoid",
            }
        }
        validate_response(client, "post", url, json_headers, data, expected_resp_code)


def test_document_request_remove_provider(client, json_headers, testdata, users):
    """Test remove provider from Document Request permissions."""

    tests = [
        ("patron1", "dreq-1", 403),
        ("librarian", "dreq-1", 202),
        ("admin", "dreq-1", 202),
        ("anonymous", "dreq-1", 401),
    ]

    for user, res_id, expected_resp_code in tests:
        user_login(client, user, users)
        url = url_for(
            "invenio_app_ils_document_requests.dreqid_provider",
            pid_value=res_id,
        )
        validate_response(client, "delete", url, json_headers, None, expected_resp_code)


def test_document_request_accept(client, json_headers, testdata, users):
    """Test Document Request permissions to accept request."""

    tests = [
        ("patron1", "dreq-1", 403),
        ("librarian", "dreq-5", 202),
        ("admin", "dreq-6", 202),
        ("anonymous", "dreq-1", 401),
        ("admin", "dreq-2", 400),
    ]

    for user, res_id, expected_resp_code in tests:
        user_login(client, user, users)
        url = url_for(
            "invenio_app_ils_document_requests.dreqid_accept", pid_value=res_id
        )
        data = {"state": "ACCEPTED"}
        validate_response(client, "post", url, json_headers, data, expected_resp_code)


def test_document_request_decline(client, json_headers, testdata, users):
    """Test Document Request permissions to decline request."""

    tests = [
        ("patron1", "dreq-1", 202),
        ("librarian", "dreq-5", 202),
        ("admin", "dreq-6", 202),
        ("anonymous", "dreq-1", 401),
        ("admin", "dreq-2", 400),
    ]

    for user, res_id, expected_resp_code in tests:
        user_login(client, user, users)
        url = url_for(
            "invenio_app_ils_document_requests.dreqid_decline",
            pid_value=res_id,
        )
        data = {"decline_reason": "USER_CANCEL"}
        validate_response(client, "post", url, json_headers, data, expected_resp_code)



# Copyright (C) 2021 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Database model for notifications."""

from datetime import datetime

from invenio_db import db


class NotificationsLogs(db.Model):
    """Keep track of sent notifications."""

    __tablename__ = "notifications_logs"

    id = db.Column(db.Integer, primary_key=True)

    is_manually_triggered = db.Column(db.Boolean, nullable=False)
    """If it was triggered manually or automatically."""

    recipient_user_id = db.Column(db.Integer, nullable=False)
    """The recipient user id."""

    triggered_on = db.Column(
        db.DateTime, nullable=False, default=lambda: datetime.now()
    )
    """Datetime the notification was sent."""

    send_log = db.Column(db.String, nullable=False)
    """Info from success/error celery callbacks."""

    message_id = db.Column(db.String, nullable=False)
    """The UUID of them notification message."""

    action = db.Column(db.String)
    """Action that triggered the notification, e.g. checkout (optional)."""

    pid_type = db.Column(db.String)
    """The pid type of the related record (optional)."""

    pid_value = db.Column(db.String)
    """The pid value of the related record (optional)."""

    @classmethod
    def create(cls, data):
        """Create a new log entry."""
        log = cls(**data)
        db.session.add(log)
        db.session.commit()
        return log



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test patrons endpoint."""

import json

from flask import url_for

from tests.helpers import user_login

_LIST_ENDPOINT = "invenio_records_rest.patid_list"


def _patron_loans_request(client, json_headers, document_pid):
    """Perform a patron loans request."""
    response = client.get(
        url_for(
            "invenio_app_ils_patrons.get_user_information",
            document_pid=document_pid,
        ),
        headers=json_headers,
    )
    return json.loads(response.data.decode("utf-8"))


def _assert_patron_loans_information(client, json_headers, document_pid, expect):
    """Assert patron loan information."""
    resp = _patron_loans_request(client, json_headers, document_pid)
    assert resp == expect


def _assert_error_when_not_authenticated(client, json_headers, document_pid):
    """Assert user is not logged in."""
    resp = _patron_loans_request(client, json_headers, document_pid)
    assert resp["status"] == 401


def test_patrons_permissions(client, testdata, json_headers, users):
    """Test patron endpoints permissions."""
    tests = [
        ("admin", [200]),
        ("librarian", [200]),
        ("patron1", [403]),
        ("anonymous", [401]),
    ]

    def _test_list(expected_status):
        """Test get list."""
        url = url_for(_LIST_ENDPOINT)
        res = client.get(url, headers=json_headers)
        assert res.status_code in expected_status

    for username, expected_status in tests:
        user_login(client, username, users)
        _test_list(expected_status)


def test_patron_loans_information(client, json_headers, testdata_most_loaned, users):
    """Test patron loans information API endpoint."""
    _assert_error_when_not_authenticated(client, json_headers, "docid-1")
    user_login(client, "librarian", users)

    _assert_patron_loans_information(
        client,
        json_headers,
        "docid-1",
        expect={
            "has_active_loan": False,
            "is_requested": False,
            "last_loan": None,
        },
    )

    user_login(client, "patron1", users)

    _assert_patron_loans_information(
        client,
        json_headers,
        "docid-1",
        expect={
            "has_active_loan": False,
            "is_requested": True,
            "last_loan": "2019-04-02",
        },
    )

    user_login(client, "patron2", users)
    _assert_patron_loans_information(
        client,
        json_headers,
        "docid-3",
        expect={
            "has_active_loan": False,
            "is_requested": True,
            "last_loan": None,
        },
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio App ILS circulation Loan Checkout loader JSON schema."""

import arrow
from flask import current_app
from invenio_circulation.records.loaders.schemas.json import (
    DateString,
    LoanItemPIDSchemaV1,
)
from marshmallow import ValidationError, fields, post_load, validates

from invenio_app_ils.permissions import check_permission

from .base import LoanBaseSchemaV1


class LoanCheckoutSchemaV1(LoanBaseSchemaV1):
    """Loan checkout schema."""

    item_pid = fields.Nested(LoanItemPIDSchemaV1, required=True)
    start_date = DateString()
    end_date = DateString()
    force = fields.Bool(missing=False)

    @validates("force")
    def validate_force(self, value, **kwargs):
        """Validate that only librarian can perform a force checkout."""
        if value:
            # extra permission for force-checkout
            permission = current_app.config["ILS_VIEWS_PERMISSIONS_FACTORY"](
                "circulation-loan-force-checkout"
            )
            check_permission(permission)

    @post_load()
    def postload_checks(self, data, **kwargs):
        """Validate dates values."""
        if "end_date" in data and "start_date" not in data:
            raise ValidationError(
                "Start date is required when end date provided.",
                "start_date",
            )

        if "start_date" in data and "end_date" in data:
            start = arrow.get(data["start_date"]).date()
            end = arrow.get(data["end_date"]).date()
            if end < start:
                raise ValidationError(
                    {
                        "start_date": [
                            "The loan start date cannot be after the end date."
                        ],
                        "end_date": [
                            "The loan end date cannot" " be before the start date."
                        ],
                    }
                )

        return data



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test acquisition provider."""

import json

from flask import url_for

from tests.helpers import user_login

_HTTP_OK = [200, 201, 204]
PROVIDER_PID = "acq-provid-1"
PROVIDER_NAME = "A provider"
PROVIDER_TYPE = "VENDOR"
ITEM_ENDPOINT = "invenio_records_rest.provid_item"
LIST_ENDPOINT = "invenio_records_rest.provid_list"


def test_acq_providers_permissions(client, testdata, json_headers, users):
    """Test providers endpoints permissions."""
    dummy_acquisition_provider = dict(name=PROVIDER_NAME, type=PROVIDER_TYPE)
    tests = [
        ("admin", _HTTP_OK, dummy_acquisition_provider),
        ("librarian", _HTTP_OK, dummy_acquisition_provider),
        ("patron1", [403], dummy_acquisition_provider),
        ("anonymous", [401], dummy_acquisition_provider),
    ]

    def _test_list(expected_status):
        """Test get list."""
        url = url_for(LIST_ENDPOINT)
        res = client.get(url, headers=json_headers)
        assert res.status_code in expected_status

    def _test_create(expected_status, data):
        """Test record creation."""
        url = url_for(LIST_ENDPOINT)
        print("DICT", json.dumps(data))
        res = client.post(url, headers=json_headers, data=json.dumps(data))
        assert res.status_code in expected_status

        if res.status_code < 400:
            record = res.get_json()["metadata"]
            assert record["name"] == PROVIDER_NAME
            assert record["type"] == PROVIDER_TYPE
            return record["pid"]

    def _test_update(expected_status, data, pid):
        """Test record update."""
        pid_value = pid or PROVIDER_PID
        url = url_for(ITEM_ENDPOINT, pid_value=pid_value)
        res = client.put(url, headers=json_headers, data=json.dumps(data))
        assert res.status_code in expected_status
        if res.status_code < 400:
            record = res.get_json()["metadata"]
            assert record["name"] == PROVIDER_NAME

    def _test_read(expected_status, pid):
        """Test record read."""
        pid_value = pid or PROVIDER_PID
        url = url_for(ITEM_ENDPOINT, pid_value=pid_value)
        res = client.get(url, headers=json_headers)
        assert res.status_code in expected_status

    def _test_delete(expected_status, pid):
        """Test record delete."""
        pid_value = pid or PROVIDER_PID
        url = url_for(ITEM_ENDPOINT, pid_value=pid_value)
        res = client.delete(url, headers=json_headers)
        assert res.status_code in expected_status

    for username, expected_status, data in tests:
        user_login(client, username, users)
        _test_list(expected_status)
        pid = _test_create(expected_status, data)
        _test_update(expected_status, data, pid)
        _test_read(expected_status, pid)
        _test_delete(expected_status, pid)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2018-2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Helpers for tests."""

import json
import os

from flask_security import login_user, logout_user
from invenio_accounts.models import User
from invenio_accounts.testutils import login_user_via_session
from invenio_db import db
from invenio_pidstore.models import PersistentIdentifier, PIDStatus

from invenio_app_ils.items.api import Item


def load_json_from_datadir(filename):
    """Load JSON from dir."""
    _data_dir = os.path.join(os.path.dirname(__file__), "data")
    with open(os.path.join(_data_dir, filename), "r") as fp:
        return json.load(fp)


def mint_record_pid(pid_type, pid_field, record):
    """Mint the given PID for the given record."""
    PersistentIdentifier.create(
        pid_type=pid_type,
        pid_value=record[pid_field],
        object_type="rec",
        object_uuid=record.id,
        status=PIDStatus.REGISTERED,
    )
    db.session.commit()


def internal_location_ref_builder(app, item_pid):
    """Ref builder for item InternalLocation."""
    path = Item._internal_location_resolver_path
    return path.format(
        scheme=app.config["JSONSCHEMAS_URL_SCHEME"],
        host=app.config["JSONSCHEMAS_HOST"],
        item_pid=item_pid,
    )


def document_ref_builder(app, item_pid):
    """Ref builder for item Document."""
    path = Item._document_resolver_path
    return path.format(
        scheme=app.config["JSONSCHEMAS_URL_SCHEME"],
        host=app.config["JSONSCHEMAS_HOST"],
        item_pid=item_pid,
    )


def user_login(client, username, users):
    """Util function to log in user."""
    user_logout(client)
    if username != "anonymous":
        user = User.query.get(users[username].id)
        # needed for Identity/Permissions loading
        login_user(user, remember=True)
        # needed for sessions/http requests
        login_user_via_session(client, email=user.email)

        return user


def user_logout(client):
    """Util function to log out user."""
    with client.session_transaction() as sess:
        from flask import g

        if "user_id" in sess:
            del sess["user_id"]
            del sess["_user_id"]
            logout_user()

        if "_login_user" in g:
            del g._login_user


def validate_data(key, expected_output, res):
    """Util function for testing output data."""
    if res.status_code != 403 and res.status_code != 401:
        data = json.loads(res.data.decode("utf-8"))["metadata"]
        assert data[key] == expected_output


def validate_response(client, req_method, url, headers, data, expected_resp_code):
    """Util function testing response code."""
    if data:
        res = getattr(client, req_method)(url, headers=headers, data=json.dumps(data))
    else:
        res = getattr(client, req_method)(url, headers=headers)
    assert expected_resp_code == res.status_code
    return res


def get_test_record(testdata, type, pid_value):
    """Fetch and return the record from the test data."""
    return [rec for rec in testdata[type] if rec["pid"] == pid_value][0]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Patrons CLI module."""

from pprint import pprint

import click
from flask.cli import with_appcontext

from invenio_app_ils.proxies import current_app_ils

from .anonymization import anonymize_patron_data, get_patron_activity


@click.group()
def patrons():
    """Patrons data CLI."""


@patrons.command()
@with_appcontext
def index():
    """Reindex all patrons."""
    click.secho("Started reindexing patrons...", fg="green")
    n_patrons = current_app_ils.patron_indexer.reindex_patrons()
    click.secho("Reindexed {} patrons".format(n_patrons), fg="green")


@patrons.command()
@click.option("--patron-pid", help="Give patron pid.")
@with_appcontext
def list_activity(patron_pid):
    """List patron's data and activity."""
    patron_activity = get_patron_activity(patron_pid)
    if not patron_activity:
        print("The patron with pid", patron_pid, "does not exist.")
        return

    pprint(patron_activity)


@patrons.command()
@click.option("--patron-pid", help="Give patron pid.")
@click.option(
    "--force",
    is_flag=True,
    default=False,
    help="Try to anonymize data by using the given patron PID even if "
    "the Invenio user does not exist.",
)
@with_appcontext
def anonymize(patron_pid, force):
    """Anonymize patron's data and activity."""
    if click.confirm("Are you sure you want to anonymize this patron?"):
        dropped, indices, notifications = anonymize_patron_data(patron_pid, force)
        msg = (
            "Successfully anonymized patron's activity: {dropped} rows "
            "deleted from db, {indices} records re-indexed and "
            "{notifications} notifications"
            " anonymized.".format(
                dropped=dropped, indices=indices, notifications=notifications
            )
        )
        click.secho(msg, fg="green")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test ILL borrowing requests."""

import json

from flask import url_for
from invenio_search import current_search

from tests.helpers import user_login

_HTTP_OK = [200, 201, 204]
BRWREQ_PID = "illbid-1"
ITEM_ENDPOINT = "invenio_records_rest.illbid_item"
LIST_ENDPOINT = "invenio_records_rest.illbid_list"


def test_ill_brwreqs_list_permissions(client, testdata, json_headers, users):
    """Test borrowing requests list permissions."""
    patron1_brwreq = dict(
        status="PENDING",
        document_pid="docid-1",
        patron_pid="1",
        provider_pid="ill-provid-1",
        type="PHYSICAL_COPY",
    )
    patron2_brwreq = dict(
        status="PENDING",
        document_pid="docid-1",
        patron_pid="2",
        provider_pid="ill-provid-1",
        type="PHYSICAL_COPY",
    )

    def _test_list(expected_status, pids):
        """Test get list for given pids."""
        q = " OR ".join(["pid:{}".format(pid) for pid in pids])
        list_url = url_for(LIST_ENDPOINT, q=q)
        res = client.get(list_url, headers=json_headers)
        assert res.status_code in expected_status
        return res.get_json()

    # create records
    list_url = url_for(LIST_ENDPOINT)
    user_login(client, "admin", users)
    res = client.post(list_url, headers=json_headers, data=json.dumps(patron1_brwreq))
    patron1_brwreq_pid = res.get_json()["metadata"]["pid"]

    res = client.post(list_url, headers=json_headers, data=json.dumps(patron2_brwreq))
    patron2_brwreq_pid = res.get_json()["metadata"]["pid"]
    all_pids = [patron1_brwreq_pid, patron2_brwreq_pid]
    # wait for ES
    current_search.flush_and_refresh(index="ill_borrowing_requests")

    # test results
    tests = [
        ("admin", _HTTP_OK, all_pids),
        ("librarian", _HTTP_OK, all_pids),
        ("patron1", _HTTP_OK, [patron1_brwreq_pid]),
        ("patron2", _HTTP_OK, [patron2_brwreq_pid]),
    ]
    for username, expected_status, expected_pids in tests:
        user_login(client, username, users)
        results = _test_list(expected_status, all_pids)
        assert results["hits"]["total"] == len(expected_pids)
        found_pids = [hit["metadata"]["pid"] for hit in results["hits"]["hits"]]
        assert set(expected_pids) == set(found_pids)

    # anonymous
    user_login(client, "anonymous", users)
    _test_list([401], [])


def test_ill_brwreq_details_permissions(client, testdata, json_headers, users):
    """Test borrowing requests details permissions."""
    dummy_borrowing_request = dict(
        status="PENDING",
        document_pid="docid-1",
        patron_pid="1",
        provider_pid="ill-provid-1",
        type="PHYSICAL_COPY",
    )

    def _test_create(expected_status, data, user):
        """Test record creation."""
        url = url_for(LIST_ENDPOINT)
        res = client.post(url, headers=json_headers, data=json.dumps(data))
        assert res.status_code in expected_status

        if res.status_code < 400:
            brw_req = res.get_json()["metadata"]
            assert brw_req["status"] == "PENDING"
            expected_created_by = dict(type="user_id", value=str(user.id))
            assert brw_req["created_by"] == expected_created_by
            assert not brw_req.get("updated_by")
            return brw_req["pid"]

    def _test_update(expected_status, data, pid, user):
        """Test record update."""
        pid_value = pid or BRWREQ_PID
        url = url_for(ITEM_ENDPOINT, pid_value=pid_value)
        res = client.put(url, headers=json_headers, data=json.dumps(data))
        assert res.status_code in expected_status
        if res.status_code < 400:
            expected_changed_by = dict(type="user_id", value=str(user.id))
            brw_req = res.get_json()["metadata"]
            assert brw_req["created_by"] == expected_changed_by
            assert brw_req["updated_by"] == expected_changed_by

    def _test_read(expected_status, pid):
        """Test record read."""
        pid_value = pid or BRWREQ_PID
        url = url_for(ITEM_ENDPOINT, pid_value=pid_value)
        res = client.get(url, headers=json_headers)
        assert res.status_code in expected_status

    def _test_delete(expected_status, pid):
        """Test record delete."""
        url = url_for(ITEM_ENDPOINT, pid_value=pid)
        res = client.delete(url, headers=json_headers)
        assert res.status_code in expected_status

    # create/update
    tests = [
        ("anonymous", [401], dummy_borrowing_request),
        ("patron2", [403], dummy_borrowing_request),
        ("patron1", [403], dummy_borrowing_request),
        ("librarian", _HTTP_OK, dummy_borrowing_request),
        ("admin", _HTTP_OK, dummy_borrowing_request),
    ]
    for username, expected_status, data in tests:
        user = user_login(client, username, users)
        pid = _test_create(expected_status, data, user)
        _test_update(expected_status, data, pid, user)

    # get
    tests = [
        ("anonymous", [401]),
        ("patron2", [403]),
        ("patron1", _HTTP_OK),
        ("librarian", _HTTP_OK),
        ("admin", _HTTP_OK),
    ]
    for username, expected_status in tests:
        user_login(client, username, users)
        _test_read(expected_status, BRWREQ_PID)

    # delete
    tests = [
        ("anonymous", [401]),
        ("patron2", [403]),
        ("patron1", [403]),
        ("librarian", [403]),
        ("admin", _HTTP_OK),
    ]
    for username, expected_status in tests:
        user_login(client, username, users)
        _test_delete(expected_status, BRWREQ_PID)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""ILS items schemas."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""EItems schema for marshmallow loader."""

from invenio_records_rest.schemas import RecordMetadataSchemaJSONV1
from marshmallow import EXCLUDE, Schema, fields, pre_load, validate

from invenio_app_ils.eitems.api import EItem
from invenio_app_ils.records.loaders.schemas.changed_by import (
    ChangedBySchema,
    set_changed_by,
)
from invenio_app_ils.records.loaders.schemas.identifiers import IdentifierSchema


class URLSchema(Schema):
    """URL schema."""

    class Meta:
        """Meta attributes for the schema."""

        unknown = EXCLUDE

    description = fields.Str()
    login_required = fields.Bool(missing=True)
    value = fields.URL(required=True)


class FileSchema(Schema):
    """File schema."""

    class Meta:
        """Meta attributes for the schema."""

        unknown = EXCLUDE

    bucket = fields.Str()
    checksum = fields.Str()
    file_id = fields.Str()
    key = fields.Str()
    size = fields.Int()
    version_id = fields.Str()


class EItemSchemaV1(RecordMetadataSchemaJSONV1):
    """EItem schema."""

    class Meta:
        """Meta attributes for the schema."""

        unknown = EXCLUDE

    bucket_id = fields.Str()
    created_by = fields.Nested(ChangedBySchema)
    description = fields.Str()
    document_pid = fields.Str(required=True)
    eitem_type = fields.Str(required=True, validate=validate.OneOf(EItem.EITEM_TYPES))
    files = fields.List(fields.Nested(FileSchema))
    identifiers = fields.List(fields.Nested(IdentifierSchema))
    internal_notes = fields.Str()
    open_access = fields.Bool(missing=True)
    source = fields.Str()
    urls = fields.List(fields.Nested(URLSchema))

    @pre_load
    def set_changed_by(self, data, **kwargs):
        """Automatically set `created_by` and `updated_by`."""
        record = self.context.get("record")
        return set_changed_by(data, record)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Circulation tasks."""

from copy import deepcopy

from celery import shared_task
from celery.utils.log import get_task_logger
from flask import current_app
from invenio_circulation.api import Loan
from invenio_circulation.proxies import current_circulation
from invenio_db import db

from invenio_app_ils.circulation.search import get_all_expired_loans

celery_logger = get_task_logger(__name__)


@shared_task
def cancel_expired_loan_requests():
    """Cancel loan requests after expiration date has passed."""
    SystemAgent = current_app.config["ILS_PATRON_SYSTEM_AGENT_CLASS"]
    system_agent_id = str(SystemAgent.id)

    for hit in get_all_expired_loans().scan():
        loan = Loan.get_record_by_pid(hit.pid)
        duration_days = current_app.config["ILS_CIRCULATION_LOAN_REQUEST_DURATION_DAYS"]
        params = deepcopy(loan)
        params.update(
            dict(
                cancel_reason="The loan request has been automatically "
                "cancelled because {} days have passed.".format(duration_days),
                transaction_user_pid=system_agent_id,
            )
        )
        current_circulation.circulation.trigger(loan, **dict(params, trigger="cancel"))

        loan.commit()
        db.session.commit()
        current_circulation.loan_indexer().index(loan)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 CERN.
#
# invenio-app-ils is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Invenio acquisition module for Integrated Library System app."""

from flask import current_app
from invenio_rest.errors import RESTException
from werkzeug.utils import cached_property

from invenio_app_ils.ext import handle_rest_exceptions

from . import config
from .api import ORDER_PID_TYPE


class _InvenioIlsAcquisitionState(object):
    """Invenio ILS acquisition app."""

    def __init__(self, app):
        """Initialize state."""
        self.app = app

    def record_class_by_pid_type(self, pid_type):
        """Return a record class by pid type."""
        endpoints = current_app.config.get("RECORDS_REST_ENDPOINTS", [])
        return endpoints[pid_type]["record_class"]

    def search_class_by_pid_type(self, pid_type):
        """Return a search class by pid type."""
        endpoints = current_app.config.get("RECORDS_REST_ENDPOINTS", [])
        return endpoints[pid_type]["search_class"]

    def indexer_by_pid_type(self, pid_type):
        """Return an indexer instance by pid type."""
        endpoints = current_app.config.get("RECORDS_REST_ENDPOINTS", [])
        return endpoints[pid_type]["indexer_class"]()

    @cached_property
    def order_record_cls(self):
        """Return the Order record class."""
        return self.record_class_by_pid_type(ORDER_PID_TYPE)

    @cached_property
    def order_indexer(self):
        """Return an Order indexer instance."""
        return self.indexer_by_pid_type(ORDER_PID_TYPE)

    @cached_property
    def order_search_cls(self):
        """Return the Order search class."""
        return self.search_class_by_pid_type(ORDER_PID_TYPE)


class InvenioIlsAcquisition(object):
    """Invenio ILS Acquisition app."""

    def __init__(self, app=None):
        """Extension initialization."""
        if app:
            self.app = app
            self.init_app(app)
            app.errorhandler(RESTException)(handle_rest_exceptions)

    def init_app(self, app):
        """Flask application initialization."""
        self.init_config(app)
        app.extensions["invenio-ils-acq"] = _InvenioIlsAcquisitionState(app)

    def init_config(self, app):
        """Initialize configuration."""
        acq_endpoints = getattr(config, "RECORDS_REST_ENDPOINTS")
        app.config.setdefault("RECORDS_REST_ENDPOINTS", {})
        app.config["RECORDS_REST_ENDPOINTS"].update(acq_endpoints)

        acq_sort_options = getattr(config, "RECORDS_REST_SORT_OPTIONS")
        app.config.setdefault("RECORDS_REST_SORT_OPTIONS", {})
        app.config["RECORDS_REST_SORT_OPTIONS"].update(acq_sort_options)

        acq_facets = getattr(config, "RECORDS_REST_FACETS")
        app.config.setdefault("RECORDS_REST_FACETS", {})
        app.config["RECORDS_REST_FACETS"].update(acq_facets)

        for k in dir(config):
            if k.startswith("ILS_ACQ_"):
                app.config.setdefault(k, getattr(config, k))
